% *** ARD-VB Algorithm ***
% *high-dimensional VB algorithm for sparse prior and homogenous noise
%  to solve set of equations m = Ax + e, where x is sparse vector of unknowns
% *see paper 'Comparison of sparse Bayesian regularization methods
%  for high-dimensional inverse problem'
% *'visualisation of the reconstruction' is prepared for case,
%  where x is n*n image - a modification is easy and possible
%
% % written by Jakub Sevcik and Vaclav Smidl 2016

% Load the measurement matrix (A) and the vector (m)
% *a training dataset is downloadable from http://www.fips.fi/dataset.php
load Data328 A m
m = m(:);

% % ARD-VB Algorithm
% inicialization
iterations = 100;

[M,N] = size(A);
n = sqrt(N);

alfa = 10*ones(N,1);
gammaA0 = 1e-10*ones(N,1);
deltaA0 = 1e-10*ones(N,1);
beta = 1;
gammab0 = 1e-10;
deltab0 = 1e-10;

% calculate dAtA = diag(A'*A) for faster computation
dAtA = NaN(size(A,2),1);
for i = 1:size(A,2)
    dAtA(i) = A(:,i)'*A(:,i);
end

% iterative section
for iter = 1:1:iterations
    diSb = beta*dAtA + alfa; % diSb = diag(beta*AtA+diag(alfa))
    dSb = 1./diSb;
    
    fun = @(x) beta*(A'*(A*x)) + (alfa.*x);
    b = A'*m*beta;
    [x,flag] = pcg(fun,b,[],50);
    %flags(iter) = flag;
    %X(iter) = x;
    %x(x<0) = 0.0*x(x<0);
    
    gammaA = gammaA0 + 0.5;
    deltaA = deltaA0 + 0.5*x.^2 + 0.5*dSb;
    alfa = gammaA./deltaA;
    
    gammab = gammab0 + 0.5*M;
    deltab = deltab0 + 0.5*m'*m + (-m' + 0.5*(x'*A'))*(A*x) + 0.5*dAtA'*dSb;
    beta = gammab/deltab;
    
    if mod(iter,10) == 0
        disp([iter, iterations])
    end
end

% % visualisation of the reconstruction
figure
imagesc(reshape(x,n,n))
colormap gray
axis square
axis off
title(['ARD-VB Algorithm with ',num2str(iterations),' iterations'])
